@@ -1,5 +1,5 @@
 /* target operations in the Linux kernel mode
- * Copyright (C) 2005-2016 Red Hat Inc.
+ * Copyright (C) 2005-2019 Red Hat Inc.
  * Copyright (C) 2005-2007 Intel Corporation.
  * Copyright (C) 2007 Quentin Barnes.
  *
@@ -30,10 +30,31 @@
 #endif
 
 
-#define k_fetch_register(regno) \
-  pt_regs_fetch_register(c->kregs, regno)
-#define k_store_register(regno, value) \
-  pt_regs_store_register(c->kregs, regno, value)
+#define check_fetch_register(regs,regno,maxregno,fn) ({		            \
+  if ((regs) == 0 || (regno) < 0 || (regno) > (maxregno)) {		    \
+        snprintf(c->error_buffer, sizeof(c->error_buffer),		    \
+		 STAP_MSG_LOC2C_04); 					    \
+    c->last_error = c->error_buffer;					    \
+    goto deref_fault;							    \
+  }                  							    \
+  fn(regs, regno);	   						    \
+})
+    
+#define check_store_register(regs,regno,maxregno,value,fn) do {		    \
+  if ((regs) == 0 || (regno) < 0 || (regno) > (maxregno)) {		    \
+        snprintf(c->error_buffer, sizeof(c->error_buffer),		    \
+		 STAP_MSG_LOC2C_04); 					    \
+    c->last_error = c->error_buffer;					    \
+    goto deref_fault;							    \
+  }                  							    \
+  fn(regs, regno, value);                                                   \
+} while(0)
+
+
+#define k_fetch_register(regno) check_fetch_register(c->kregs,regno,pt_regs_maxno,pt_regs_fetch_register)
+
+#define k_store_register(regno,value) check_store_register(c->kregs,regno,pt_regs_maxno,value,pt_regs_store_register)
+    
 
 
 /* PR 10601: user-space (user_regset) register access.
@@ -258,15 +279,22 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
 }
 
 
+
 #if defined (__i386__)
 
-#define u_fetch_register(regno) ursl_fetch32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno)
-#define u_store_register(regno,value) ursl_store32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno, value)
+#define uu_fetch_register(_regs,regno) ursl_fetch32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno)
+#define uu_store_register(_regs,regno,value) ursl_store32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno, value)
+
+#define u_fetch_register(regno) check_fetch_register(c->uregs,regno,ARRAY_SIZE(url_i386),uu_fetch_register)
+#define u_store_register(regno,value) check_store_register(c->uregs,regno,ARRAY_SIZE(url_i386),value,uu_store_register)
 
 #elif defined (__x86_64__)
 
-#define u_fetch_register(regno) (_stp_is_compat_task() ? ursl_fetch32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno) : ursl_fetch64(url_x86_64, ARRAY_SIZE(url_x86_64), EM_X86_64, regno))
-#define u_store_register(regno,value)  (_stp_is_compat_task() ? ursl_store32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno, value) : ursl_store64(url_x86_64, ARRAY_SIZE(url_x86_64), EM_X86_64, regno, value))
+#define uu_fetch_register(_regs,regno) (_stp_is_compat_task() ? ursl_fetch32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno) : ursl_fetch64(url_x86_64, ARRAY_SIZE(url_x86_64), EM_X86_64, regno))
+#define uu_store_register(_regs,regno,value)  (_stp_is_compat_task() ? ursl_store32(url_i386, ARRAY_SIZE(url_i386), EM_386, regno, value) : ursl_store64(url_x86_64, ARRAY_SIZE(url_x86_64), EM_X86_64, regno, value))
+
+#define u_fetch_register(regno) check_fetch_register(c->uregs,regno,_stp_is_compat_task()?ARRAY_SIZE(url_i386):ARRAY_SIZE(url_x86_64),uu_fetch_register)
+#define u_store_register(regno,value) check_store_register(c->uregs,regno,_stp_is_compat_task()?ARRAY_SIZE(url_i386):ARRAY_SIZE(url_x86_64),value,uu_store_register)
 
 #endif
 
@@ -275,17 +303,17 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
 /* Downgrade to pt_dwarf_register access. */
 
 #define u_store_register(regno, value) \
-  pt_regs_store_register(c->uregs, regno, value)
+  check_store_register(c->uregs,regno,pt_regs_maxno,value,pt_regs_store_register)
 
 /* If we're in a 32/31-bit task in a 64-bit kernel, we need to emulate
  * 32-bitness by masking the output of pt_regs_fetch_register() */
 #ifndef CONFIG_COMPAT
 #define u_fetch_register(regno) \
-  pt_regs_fetch_register(c->uregs, regno)
+  check_fetch_register(c->uregs,regno,pt_regs_maxno,pt_regs_fetch_register)
+
 #else
 #define u_fetch_register(regno) \
-  (_stp_is_compat_task() ? (0xffffffff & pt_regs_fetch_register(c->uregs, regno)) \
-                         : pt_regs_fetch_register(c->uregs, regno))
+  check_fetch_register(c->uregs,regno,pt_regs_maxno,pt_regs_fetch_register) & (_stp_is_compat_task() ? 0xffffffff : ~(int64_t)0)
 #endif
 
 #endif /* STAPCONF_REGSET */
@@ -348,60 +376,82 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * that the kernel doesn't pagefault while reading.
  */
 
-#define __stp_deref_nocheck(value, size, addr)				      \
-  ({									      \
-    int _err = -EFAULT;							      \
-    switch (size)							      \
-      {									      \
-      case 1: { u8 _b; _err = __stp_get_user(_b, ((u8 *)(uintptr_t)(addr)));  \
-		(value) = _b; }						      \
-	break;								      \
-      case 2: { u16 _w; _err = __stp_get_user(_w, ((u16 *)(uintptr_t)(addr))); \
-		(value) = _w; }						      \
-	break;								      \
-      case 4: { u32 _l; _err = __stp_get_user(_l, ((u32 *)(uintptr_t)(addr))); \
-		(value) = _l; }						      \
-	break;								      \
-      case 8: { u64 _q; _err = __stp_get_user(_q, ((u64 *)(uintptr_t)(addr))); \
-		(value) = _q; }						      \
-	break;								      \
-      default:								      \
- 	__stp_get_user_bad();					      	      \
-      }									      \
-    _err;								      \
+static inline int __stp_deref_nocheck_(u64 *pv, size_t size, void *addr)
+{
+  u64 v = 0;
+  int r = -EFAULT;
+
+  switch (size)
+    {
+    case 1: { u8 b; r = __stp_get_user(b, (u8 *)addr); v = b; } break;
+    case 2: { u16 w; r = __stp_get_user(w, (u16 *)addr); v = w; } break;
+    case 4: { u32 l; r = __stp_get_user(l, (u32 *)addr); v = l; } break;
+#if defined(__i386__) || defined(__arm__)
+    /* x86 and arm can't do 8-byte get_user, so we have to split it */
+    case 8: { union { u32 l[2]; u64 ll; } val;
+	      r = __stp_get_user(val.l[0], &((u32 *)addr)[0]);
+	      if (! r)
+		  r = __stp_get_user(val.l[1], &((u32 *)addr)[1]);
+	      if (! r)
+		  v = val.ll;
+	    } break;
+#else
+    case 8: { r = __stp_get_user(v, (u64 *)addr); } break;
+#endif
+    }
+
+  *pv = v;
+  return r;
+}
+
+#define __stp_deref_nocheck(value, size, addr)				\
+  ({									\
+    u64 _v = 0; int _e = -EFAULT;					\
+    switch (size)							\
+      {									\
+      case 1: case 2: case 4: case 8:					\
+	_e = __stp_deref_nocheck_(&_v, (size), (void *)(uintptr_t)(addr)); \
+	(value) = _v;							\
+	break;								\
+      default:								\
+	__stp_get_user_bad();						\
+      }									\
+    _e;									\
   })
 
 
 /* 
- * _stp_deref(): safely read a simple type from memory
+ * _stp_lookup_bad_addr(): safely verify an address
  *
- * size: number of bytes to read
- * addr: address to read from
- * seg: memory segment to use, either kernel (KERN_DS) or user
+ * type: memory access type (either VERIFY_READ or VERIFY_WRITE)
+ * size: number of bytes to verify
+ * addr: address to verify
+ * seg: memory segment to use, either kernel (KERNEL_DS) or user
  * (USER_DS)
  * 
- * The macro returns the value read. If this macro gets an error while
- * trying to read memory, a DEREF_FAULT will occur. Note that the
- * kernel will not pagefault when trying to read the memory.
+ * The macro returns 0 if the address is valid, non-zero otherwise.
+ * Note that the kernel will not pagefault when trying to verify the
+ * memory. Also note that no DEREF_FAULT will occur when verifying the
+ * memory.
  */
 
-#define _stp_deref(size, addr, seg)                                           \
-  ({									      \
-    int _bad = 0;							      \
-    intptr_t _v = 0;							      \
-    mm_segment_t _oldfs = get_fs();                                           \
-    set_fs(seg);                                                              \
-    pagefault_disable();                                                      \
-    if (lookup_bad_addr(VERIFY_READ, (const unsigned long)(addr), (size)))    \
-      _bad = -EFAULT;                                                         \
-    else                                                                      \
-      _bad = __stp_deref_nocheck(_v, (size), (addr));			      \
-    pagefault_enable();                                                       \
-    set_fs(_oldfs);                                                           \
-    if (_bad)                                                                 \
-      DEREF_FAULT(addr);						      \
-    _v;									      \
-  })
+static inline int _stp_lookup_bad_addr_(int type, size_t size,
+                                        uintptr_t addr, mm_segment_t seg)
+{
+  int bad;
+  mm_segment_t oldfs = get_fs();
+
+  set_fs(seg);
+  pagefault_disable();
+  bad = lookup_bad_addr(type, addr, size);
+  pagefault_enable();
+  set_fs(oldfs);
+
+  return bad;
+}
+
+#define _stp_lookup_bad_addr(type, size, addr, seg) \
+  _stp_lookup_bad_addr_((type), (size), (uintptr_t)(addr), (seg))
 
 
 /* 
@@ -411,7 +461,7 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * value: read the simple type into this variable
  * size: number of bytes to read
  * addr: address to read from
- * seg: memory segment to use, either kernel (KERN_DS) or user
+ * seg: memory segment to use, either kernel (KERNEL_DS) or user
  * (USER_DS)
  * 
  * If this macro gets an error while trying to read memory, nonzero is
@@ -419,20 +469,69 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * pagefault when trying to read the memory.
  */
 
-#define _stp_deref_nofault(value, size, addr, seg)			      \
-  ({									      \
-    int _bad = 0;							      \
-    intptr_t _v = 0;							      \
-    mm_segment_t _oldfs = get_fs();                                           \
-    set_fs(seg);                                                              \
-    pagefault_disable();                                                      \
-    if (lookup_bad_addr(VERIFY_READ, (const unsigned long)(addr), (size)))    \
-      _bad = -EFAULT;							      \
-    else                                                                      \
-      _bad = __stp_deref_nocheck((value), (size), (addr));		      \
-    pagefault_enable();                                                       \
-    set_fs(_oldfs);                                                           \
-    _bad;								      \
+static inline int _stp_deref_nofault_(u64 *pv, size_t size, void *addr,
+				      mm_segment_t seg)
+{
+  int r = -EFAULT;
+  mm_segment_t oldfs = get_fs();
+
+  set_fs(seg);
+  pagefault_disable();
+  if (lookup_bad_addr(VERIFY_READ, (uintptr_t)addr, size))
+    r = -EFAULT;
+  else
+    r = __stp_deref_nocheck_(pv, size, addr);
+  pagefault_enable();
+  set_fs(oldfs);
+
+  return r;
+}
+
+#define _stp_deref_nofault(value, size, addr, seg)			\
+  ({									\
+    u64 _v = 0; int _e = -EFAULT;					\
+    switch (size)							\
+      {									\
+      case 1: case 2: case 4: case 8:					\
+	_e = _stp_deref_nofault_					\
+		(&_v, (size), (void *)(uintptr_t)(addr), (seg));	\
+	break;								\
+      default:								\
+	__stp_get_user_bad();						\
+      }									\
+    (value) = _v;							\
+    _e;									\
+  })
+
+
+/* 
+ * _stp_deref(): safely read a simple type from memory
+ *
+ * size: number of bytes to read
+ * addr: address to read from
+ * seg: memory segment to use, either kernel (KERNEL_DS) or user
+ * (USER_DS)
+ * 
+ * The macro returns the value read. If this macro gets an error while
+ * trying to read memory, a DEREF_FAULT will occur. Note that the
+ * kernel will not pagefault when trying to read the memory.
+ */
+
+#define _stp_deref(size, addr, seg)					\
+  ({									\
+    u64 _v = 0; int _e = -EFAULT;					\
+    switch (size)							\
+      {									\
+      case 1: case 2: case 4: case 8:					\
+	_e = _stp_deref_nofault_					\
+		(&_v, (size), (void *)(uintptr_t)(addr), (seg));	\
+	break;								\
+      default:								\
+	__stp_get_user_bad();						\
+      }									\
+    if (_e)								\
+      DEREF_FAULT(addr);						\
+    _v;									\
   })
 
 
@@ -451,23 +550,42 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * that the kernel doesn't pagefault while writing.
  */
 
-#define __stp_store_deref_nocheck(size, addr, value)			      \
-  ({									      \
-    int _err = -EFAULT;							      \
-    switch (size)							      \
-      {									      \
-      case 1: _err = __stp_put_user(((u8)(value)), ((u8 *)(uintptr_t)(addr)));\
-        break;								      \
-      case 2: _err = __stp_put_user(((u16)(value)), ((u16 *)(uintptr_t)(addr)));\
-        break;								      \
-      case 4: _err = __stp_put_user(((u32)(value)), ((u32 *)(uintptr_t)(addr)));\
-        break;								      \
-      case 8: _err = __stp_put_user(((u64)(value)), ((u64 *)(uintptr_t)(addr)));\
-        break;								      \
-      default:								      \
-	__stp_put_user_bad();						      \
-      }									      \
-    _err;								      \
+static inline int __stp_store_deref_nocheck_(size_t size, void *addr, u64 v)
+{
+  int r;
+  switch (size)
+    {
+    case 1: r = __stp_put_user((u8)v, (u8 *)addr); break;
+    case 2: r = __stp_put_user((u16)v, (u16 *)addr); break;
+    case 4: r = __stp_put_user((u32)v, (u32 *)addr); break;
+#if defined(__i386__) || defined(__arm__)
+    /* x86 and arm can't do 8-byte put_user, so we have to split it */
+    default: { union { u32 l[2]; u64 ll; } val;
+	       val.ll = v;
+	       r = __stp_put_user(val.l[0], &((u32 *)addr)[0]);
+	       if (! r)
+		   r = __stp_put_user(val.l[1], &((u32 *)addr)[1]);
+	     } break;
+#else
+    default: r = __stp_put_user(v, (u64 *)addr); break;
+#endif
+    }
+  return r;
+}
+
+#define __stp_store_deref_nocheck(size, addr, value)			\
+  ({									\
+    int _e = -EFAULT;							\
+    switch (size)							\
+      {									\
+      case 1: case 2: case 4: case 8:					\
+	_e = __stp_store_deref_nocheck_					\
+		((size), (void *)(uintptr_t)(addr), (value));		\
+	break;								\
+      default:								\
+        __stp_put_user_bad();						\
+      }									\
+    _e;									\
   })
 
 
@@ -477,7 +595,7 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * size: number of bytes to write
  * addr: address to write to
  * value: read the simple type from this variable
- * seg: memory segment to use, either kernel (KERN_DS) or user
+ * seg: memory segment to use, either kernel (KERNEL_DS) or user
  * (USER_DS)
  * 
  * The macro has no return value. If this macro gets an error while
@@ -485,20 +603,39 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * kernel will not pagefault when trying to write the memory.
  */
 
-#define _stp_store_deref(size, addr, value, seg)                              \
-  ({									      \
-    int _bad = 0;							      \
-    mm_segment_t _oldfs = get_fs();                                           \
-    set_fs(seg);                                                              \
-    pagefault_disable();                                                      \
-    if (lookup_bad_addr(VERIFY_WRITE, (const unsigned long)(addr), (size)))   \
-      _bad = -EFAULT;                                                         \
-    else                                                                      \
-      _bad = __stp_store_deref_nocheck((size), (addr), (value));	      \
-    pagefault_enable();                                                       \
-    set_fs(_oldfs);                                                           \
-    if (_bad)								      \
-      STORE_DEREF_FAULT(addr);						      \
+static inline int _stp_store_deref_(size_t size, void *addr, u64 v,
+				    mm_segment_t seg)
+{
+  int r;
+  mm_segment_t oldfs = get_fs();
+
+  set_fs(seg);
+  pagefault_disable();
+  if (lookup_bad_addr(VERIFY_READ, (uintptr_t)addr, size))
+    r = -EFAULT;
+  else
+    r = __stp_store_deref_nocheck_(size, addr, v);
+  pagefault_enable();
+  set_fs(oldfs);
+
+  return r;
+}
+
+#define _stp_store_deref(size, addr, value, seg)			\
+  ({									\
+    int _e = -EFAULT;							\
+    switch (size)							\
+      {									\
+      case 1: case 2: case 4: case 8:					\
+	_e = _stp_store_deref_						\
+		((size), (void *)(uintptr_t)(addr), (value), (seg));	\
+	break;								\
+      default:								\
+        __stp_put_user_bad();						\
+      }									\
+    if (_e)								\
+      STORE_DEREF_FAULT(addr);						\
+    (void)0;								\
   })
 
 
@@ -520,26 +657,32 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
 #define store_uderef(s,a,v) _stp_store_deref(s,a,v,USER_DS)
 #endif
 
-#if defined (__i386__) || defined (__arm__)
+#ifndef CONFIG_64BIT
 
-/* x86 and arm can't do 8-byte put/get_user_asm, so we have to split it */
+/* The kderef/uderef macros (which is what Xderef gets set to), alway
+ * returns a 64-bit value. This causes a problem on a 32-bit system
+ * when we want to cast the 64-bit value to a 32-bit pointer - gcc
+ * gives a "cast to pointer from integer of different size" error. So,
+ * we'll cast it to a u32 before doing the final cast to the actual
+ * type. */
 
-#define __Xread(ptr, Xderef)				\
-  ((sizeof(*(ptr)) == 8) ?				\
-       *(typeof(ptr))&(u32[2]) {			\
-	 (u32) Xderef(4, &((u32 *)(ptr))[0]),		\
-	 (u32) Xderef(4, &((u32 *)(ptr))[1]) }		\
-     : (typeof(*(ptr))) Xderef(sizeof(*(ptr)), (ptr)))
+#define __Xread(ptr, Xderef)						\
+  ((sizeof(*(ptr)) == 8)						\
+   ? *(typeof(ptr))&(u64) { Xderef(sizeof(*(ptr)), (ptr)) }		\
+   : *(typeof(ptr))&(u32) { (u32) Xderef(sizeof(*(ptr)), (ptr)) } )
 
-#define __Xwrite(ptr, value, store_Xderef)				     \
-  ({									     \
-    if (sizeof(*(ptr)) == 8) {						     \
-      union { typeof(*(ptr)) v; u32 l[2]; } _kw;			     \
-      _kw.v = (typeof(*(ptr)))(value);					     \
-      store_Xderef(4, &((u32 *)(ptr))[0], _kw.l[0]);			     \
-      store_Xderef(4, &((u32 *)(ptr))[1], _kw.l[1]);			     \
-    } else								     \
-      store_Xderef(sizeof(*(ptr)), (ptr), (long)(typeof(*(ptr)))(value));    \
+/* For __Xwrite, we need to handle the case where 'value' is a pointer
+ * and avoid the "cast from pointer to integer of different size" gcc
+ * errors. */
+
+#define __Xwrite(ptr, value, store_Xderef)				  \
+  ({									  \
+    if (sizeof(*(ptr)) == 8) {						  \
+      union { typeof(*(ptr)) v; u64 l; } _kw;				  \
+      _kw.v = (typeof(*(ptr)))(value);					  \
+      store_Xderef(8, (ptr), _kw.l);					  \
+    } else								  \
+      store_Xderef(sizeof(*(ptr)), (ptr), (long)(typeof(*(ptr)))(value)); \
   })
 
 #else
@@ -567,80 +710,103 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * '\0' byte or append a '\0' byte.
  */
 
-#define kderef_buffer(dst, addr, maxbytes)				      \
-  ({									      \
-    uintptr_t _addr;							      \
-    size_t _len;							      \
-    unsigned char _c;							      \
-    char *_d = (dst);							      \
-    int _bad = 0;							      \
-    mm_segment_t _oldfs = get_fs();                                           \
-    set_fs(KERNEL_DS);							      \
-    pagefault_disable();                                                      \
-    if (lookup_bad_addr(VERIFY_READ, (const unsigned long)(addr), (maxbytes))) \
-      _bad = 1;                                                               \
-    else                                                                      \
-      for (_len = (maxbytes), _addr = (uintptr_t)(addr);		      \
-	   _len > 1; --_len, ++_addr)					      \
-        {								      \
-	  _bad = __stp_deref_nocheck(_c, 1, _addr);			      \
-	  if (_bad)							      \
-	    break;							      \
-	  else if (_d)							      \
-	    *_d++ = _c;							      \
-	}								      \
-    pagefault_enable();                                                       \
-    set_fs(_oldfs);                                                           \
-    if (_bad)                                                                 \
-      DEREF_FAULT(addr);						      \
-    (dst);								      \
+static inline char *kderef_buffer_(char *dst, void *addr, size_t len)
+{
+  int err = 0;
+  size_t i;
+  mm_segment_t oldfs = get_fs();
+
+  set_fs(KERNEL_DS);
+  pagefault_disable();
+  if (lookup_bad_addr(VERIFY_READ, (uintptr_t)addr, len))
+    err = 1;
+  else
+    for (i = 0; i < len; ++i)
+      {
+	u8 v;
+	err = __stp_get_user(v, (u8 *)addr + i);
+	if (err)
+	  break;
+	if (dst)
+	  *dst++ = v;
+      }
+  pagefault_enable();
+  set_fs(oldfs);
+
+  return err ? (char *)-1 : dst;
+}
+
+#define kderef_buffer(dst, addr, maxbytes)				\
+  ({									\
+    char *_r = kderef_buffer_((dst), (void *)(uintptr_t)(addr), (maxbytes)); \
+    if (_r == (char *)-1)						\
+      DEREF_FAULT(addr);						\
+    _r;									\
   })
 
-/* The following is for kernel strings, see the uconversions.stp
-   tapset for user_string functions. */
+/* 
+ * _stp_deref_string_nofault(): safely read a string from memory
+ * without a DEREF_FAULT on error
+ *
+ * dst: read the string into this address
+ * addr: address to read from
+ * len: maximum number of bytes to store in dst, including the trailing NUL
+ * seg: memory segment to use, either kernel (KERNEL_DS) or user
+ * (USER_DS)
+ * 
+ * If this function gets an error while trying to read memory, -EFAULT
+ * is returned. On success, the number of bytes copied is returned
+ * (not including the trailing NULL). Note that the kernel will not
+ * pagefault when trying to read the string.
+ */
 
+static inline long _stp_deref_string_nofault(char *dst, const char *addr,
+					     size_t len, mm_segment_t seg)
+{
+  int err = 0;
+  size_t i = 0;
+  mm_segment_t oldfs = get_fs();
 
-#define kderef_string(dst, addr, maxbytes)				      \
-  ({									      \
-    uintptr_t _addr;							      \
-    size_t _len;							      \
-    unsigned char _c;							      \
-    char *_d = (dst);							      \
-    int _bad = 0;							      \
-    mm_segment_t _oldfs = get_fs();                                           \
-    set_fs(KERNEL_DS);							      \
-    pagefault_disable();                                                      \
-    if (lookup_bad_addr(VERIFY_READ, (const unsigned long)(addr), (maxbytes))) \
-      _bad = 1;                                                               \
-    else                                                                      \
-      {									      \
-	for (_len = (maxbytes), _addr = (uintptr_t)(addr);		      \
-	     _len > 1; --_len, ++_addr)					      \
-	  {								      \
-	    _bad = __stp_deref_nocheck(_c, 1, _addr);			      \
-	    if (_bad || _c == '\0')					      \
-	      break;							      \
-	    else if (_d)						      \
-	      *_d++ = _c;						      \
-	  }								      \
-	if (!_bad && _d)						      \
-	  *_d = '\0';							      \
-      }									      \
-    pagefault_enable();                                                       \
-    set_fs(_oldfs);                                                           \
-    if (_bad)                                                                 \
-      DEREF_FAULT(addr);						      \
-    (dst);								      \
+  set_fs(seg);
+  pagefault_disable();
+  if (lookup_bad_addr(VERIFY_READ, (uintptr_t)addr, len))
+    err = 1;
+  else
+    {
+      /* Reduce len by 1 to leave room for '\0' terminator. */
+      for (i = 0; i + 1 < len; ++i)
+	{
+	  u8 v;
+	  err = __stp_get_user(v, (u8 *)addr + i);
+	  if (err || v == '\0')
+	    break;
+	  if (dst)
+	    *dst++ = v;
+	}
+      if (!err && dst)
+	*dst = '\0';
+    }
+  pagefault_enable();
+  set_fs(oldfs);
+
+  return err ? -EFAULT : i;
+}
+
+#define kderef_string(dst, addr, maxbytes)				\
+  ({									\
+    long _r = _stp_deref_string_nofault((dst), (void *)(uintptr_t)(addr), (maxbytes), KERNEL_DS); \
+    if (_r < 0)								\
+      DEREF_FAULT(addr);						\
+    _r;									\
   })
 
-
 /* 
  * _stp_store_deref_string(): safely write a string to memory
  *
  * src: source string
  * addr: address to write to
  * maxbytes: maximum number of bytes to write
- * seg: memory segment to use, either kernel (KERN_DS) or user
+ * seg: memory segment to use, either kernel (KERNEL_DS) or user
  * (USER_DS)
  * 
  * The macro has no return value. If this macro gets an error while
@@ -648,33 +814,40 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * kernel will not pagefault when trying to write the memory.
  */
 
-#define _stp_store_deref_string(src, addr, maxbytes, seg)		      \
-  ({									      \
-    uintptr_t _addr;							      \
-    size_t _len;							      \
-    char *_s = (src);							      \
-    int _bad = 0;							      \
-    mm_segment_t _oldfs = get_fs();                                           \
-    set_fs(seg);                                                              \
-    pagefault_disable();                                                      \
-    if (lookup_bad_addr(VERIFY_WRITE, (const unsigned long)(addr), (maxbytes))) \
-      _bad = 1;                                                               \
-    else                                                                      \
-      {									      \
-	for (_len = (maxbytes), _addr = (uintptr_t)(addr);		      \
-	     _len > 1 && _s && *_s != '\0'; --_len, ++_addr)		      \
-	  {								      \
-	    _bad = __stp_store_deref_nocheck(1, _addr, *_s++);		      \
-	    if (_bad)					      		      \
-	      break;							      \
-	  }								      \
-	if (!_bad)							      \
-	  __stp_store_deref_nocheck(1, _addr, '\0');			      \
-      }									      \
-    pagefault_enable();                                                       \
-    set_fs(_oldfs);                                                           \
-    if (_bad)								      \
-      STORE_DEREF_FAULT(addr);						      \
+static inline int _stp_store_deref_string_(char *src, void *addr, size_t len,
+					   mm_segment_t seg)
+{
+  int err = 0;
+  size_t i;
+  mm_segment_t oldfs = get_fs();
+
+  set_fs(seg);
+  pagefault_disable();
+  if (lookup_bad_addr(VERIFY_WRITE, (uintptr_t)addr, len))
+    err = 1;
+  else if (len > 0)
+    {
+      for (i = 0; i < len - 1; ++i)
+	{
+	  err = __stp_put_user(*src++, (u8 *)addr + i);
+	  if (err)
+	    goto out;
+	}
+      err = __stp_put_user('\0', (u8 *)addr + i);
+    }
+ out:
+  pagefault_enable();
+  set_fs(oldfs);
+
+  return err;
+}
+
+#define _stp_store_deref_string(src, addr, maxbytes, seg)		\
+  ({									\
+    if (_stp_store_deref_string_					\
+	((src), (void *)(uintptr_t)(addr), (maxbytes), (seg)))		\
+      STORE_DEREF_FAULT(addr);						\
+    (void)0;								\
   })
 
 
@@ -690,7 +863,7 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * kernel will not pagefault when trying to write the memory.
  */
 
-#define store_kderef_string(src, addr, maxbytes)			      \
+#define store_kderef_string(src, addr, maxbytes)                              \
   _stp_store_deref_string((src), (addr), (maxbytes), KERNEL_DS)
 
 
@@ -706,7 +879,7 @@ static void ursl_store64 (const struct usr_regset_lut* lut,unsigned lutsize,  in
  * kernel will not pagefault when trying to write the memory.
  */
 
-#define store_uderef_string(src, addr, maxbytes)			      \
+#define store_uderef_string(src, addr, maxbytes)                              \
   _stp_store_deref_string((src), (addr), (maxbytes), USER_DS)
 
 #endif /* _LINUX_LOC2C_RUNTIME_H_ */
