@@ -100,6 +100,9 @@ struct stap_task_finder_target {
 
 /* public: */
 	pid_t pid;
+        int build_id_len;
+        uint64_t build_id_vaddr;
+        const char *build_id;
 	const char *procname;
         const char *purpose;
 	stap_task_finder_callback callback;
@@ -1264,7 +1267,7 @@ __stp_call_mmap_callbacks_for_task(struct stap_task_finder_target *tgt,
 
 	// Now allocate an array to cache vma information in.
 	if (file_based_vmas > 0)
-		vma_cache = _stp_kmalloc(sizeof(struct vma_cache_t)
+		vma_cache = _stp_vzalloc(sizeof(struct vma_cache_t)
 					 * file_based_vmas);
 	if (vma_cache != NULL) {
 		// Loop through the vmas again, and cache needed information.
@@ -1339,7 +1342,7 @@ __stp_call_mmap_callbacks_for_task(struct stap_task_finder_target *tgt,
 			}
 			vma_cache_p++;
 		}
-		_stp_kfree(vma_cache);
+		_stp_vfree(vma_cache);
 	}
 
 	_stp_kfree(mmpath_buf);
@@ -1409,11 +1412,10 @@ __stp_utrace_task_finder_target_quiesce(enum utrace_resume_action action,
 		_stp_error("utrace_set_events returned error %d on pid %d",
 			   rc, (int)tsk->pid);
 
+	/* NB make sure we run mmap callbacks before other callbacks
+	 * like 'probe process.begin' handlers so that the vma tracker
+	 * is already initialized in the latter contexts */
 
-	/* Call the callbacks.  Assume that if the thread is a
-	 * thread group leader, it is a process. */
-	__stp_call_callbacks(tgt, tsk, 1, (tsk->pid == tsk->tgid));
- 
 	/* If this is just a thread other than the thread group leader,
            don't bother inform map callback clients about its memory map,
            since they will simply duplicate each other. */
@@ -1421,6 +1423,10 @@ __stp_utrace_task_finder_target_quiesce(enum utrace_resume_action action,
 		__stp_call_mmap_callbacks_for_task(tgt, tsk);
 	}
 
+	/* Call the callbacks.  Assume that if the thread is a
+	 * thread group leader, it is a process. */
+	__stp_call_callbacks(tgt, tsk, 1, (tsk->pid == tsk->tgid));
+
 	__stp_tf_handler_end();
 	return UTRACE_RESUME;
 }
