@@ -1,7 +1,7 @@
 /*
  * utrace infrastructure interface for debugging user processes
  *
- * Copyright (C) 2006-2014 Red Hat, Inc.  All rights reserved.
+ * Copyright (C) 2006-2018 Red Hat, Inc.  All rights reserved.
  *
  * This copyrighted material is made available to anyone wishing to use,
  * modify, copy, or redistribute it subject to the terms and conditions
@@ -41,7 +41,7 @@
 // For now, disable the task_work_queue on non-RT kernels.
 // XXX: Need to decide correct CONFIG_* to check for,
 // sticking with the more conservative option for now.
-#ifdef CONFIG_PREEMPT_RT_FULL
+#if defined(CONFIG_PREEMPT_RT_FULL) || defined(CONFIG_PREEMPT_RT)
 #define STP_UTRACE_USE_TASK_WORK_QUEUE
 #endif
 /* If STP_UTRACE_USE_TASK_WORK_QUEUE is defined, reporting work
