@@ -1,6 +1,6 @@
 /* -*- linux-c -*-
  * kernel stack unwinding
- * Copyright (C) 2008-2011, 2014 Red Hat Inc.
+ * Copyright (C) 2008-2019 Red Hat Inc.
  *
  * Based on old kernel code that is
  * Copyright (C) 2002-2006 Novell, Inc.
@@ -235,7 +235,7 @@ static int parse_fde_cie(const u32 *fde, const u32 *cie,
 
 #define REG_STATE state->reg[state->stackDepth]
 
-static int advance_loc(unsigned long delta, struct unwind_state *state)
+static int advance_loc(unsigned long delta, struct uw_state *state)
 {
 	state->loc += delta * state->codeAlign;
 	dbug_unwind(1, "state->loc=%lx\n", state->loc);
@@ -244,7 +244,7 @@ static int advance_loc(unsigned long delta, struct unwind_state *state)
 
 /* Set Same or Nowhere rule for register. */
 static void set_no_state_rule(uleb128_t reg, enum item_location where,
-                              struct unwind_state *state)
+                              struct uw_state *state)
 {
 	dbug_unwind(1, "reg=%lx, where=%d\n", reg, where);
 	if (reg < ARRAY_SIZE(REG_STATE.regs)) {
@@ -254,7 +254,7 @@ static void set_no_state_rule(uleb128_t reg, enum item_location where,
 
 /* Memory or Value rule */
 static void set_offset_rule(uleb128_t reg, enum item_location where,
-                            sleb128_t svalue, struct unwind_state *state)
+                            sleb128_t svalue, struct uw_state *state)
 {
 	dbug_unwind(1, "reg=%lx, where=%d, svalue=%lx\n", reg, where, svalue);
 	if (reg < ARRAY_SIZE(REG_STATE.regs)) {
@@ -265,7 +265,7 @@ static void set_offset_rule(uleb128_t reg, enum item_location where,
 
 /* Register rule. */
 static void set_register_rule(uleb128_t reg, uleb128_t value,
-                              struct unwind_state *state)
+                              struct uw_state *state)
 {
 	dbug_unwind(1, "reg=%lx, value=%lx\n", reg, value);
 	if (reg < ARRAY_SIZE(REG_STATE.regs)) {
@@ -277,7 +277,7 @@ static void set_register_rule(uleb128_t reg, uleb128_t value,
 /* Expr or ValExpr rule. */
 static void set_expr_rule(uleb128_t reg, enum item_location where,
 			  const u8 **expr, const u8 *end,
-			  struct unwind_state *state)
+			  struct uw_state *state)
 {
 	const u8 *const start = *expr;
 	uleb128_t len = get_uleb128(expr, end);
@@ -291,12 +291,19 @@ static void set_expr_rule(uleb128_t reg, enum item_location where,
 	}
 }
 
-/* Limit the number of instructions we process. Arbitrary limit.
-   512 should be enough for anybody... */
-#define MAX_CFI 512
+/* Limit the number of instructions we process.  
+   This is intended to provide a little bit of protection
+   against CPU DoS by forcing the stap module to overly
+   large user-provided data.
+   8192 should be enough for anybody... larger than anything
+   found in e.g. fedora 29. 
+*/
+#ifndef MAX_CFI
+#define MAX_CFI 8192
+#endif
 
 static int processCFI(const u8 *start, const u8 *end, unsigned long targetLoc,
-		      signed ptrType, int user, struct unwind_state *state, int compat_task)
+		      signed ptrType, int user, struct uw_state *state, int compat_task)
 {
 	union {
 		const u8 *p8;
@@ -520,7 +527,7 @@ static int processCFI(const u8 *start, const u8 *end, unsigned long targetLoc,
 					REG_STATE.cfa.reg = value;
 					dbug_unwind(1, "DW_CFA_def_cfa reg=%ld\n", REG_STATE.cfa.reg);
 				}
-				/*nobreak */
+				/* fallthrough */
 			case DW_CFA_def_cfa_offset:
 				if (REG_STATE.cfa_is_expr != 0) {
 					_stp_warn("Unexpected DW_CFA_def_cfa_offset\n");
@@ -542,7 +549,7 @@ static int processCFI(const u8 *start, const u8 *end, unsigned long targetLoc,
 						    value, DWARF_REG_MAP(value));
 					REG_STATE.cfa.reg = value;
 				}
-				/*nobreak */
+				/* fallthrough */
 			case DW_CFA_def_cfa_offset_sf:
 				if (REG_STATE.cfa_is_expr != 0) {
 					_stp_warn("Unexpected DW_CFA_def_cfa_offset_sf\n");
@@ -724,10 +731,15 @@ adjustStartLoc (unsigned long startLoc,
   dbug_unwind(2, "adjustStartLoc=%lx, ptrType=%s, m=%s, s=%s eh=%d\n",
 	      startLoc, _stp_eh_enc_name(ptrType), m->path, s->name, is_ehframe);
   if (startLoc == 0
-      || strcmp (m->name, "kernel")  == 0
       || (strcmp (s->name, ".absolute") == 0 && !is_ehframe))
     return startLoc;
 
+  /* The .debug_frame loaded from disk is already relocated against the
+     expected load offset of the kernel, but the actual static (load)
+     address might be different (with kaslr).  */
+  if (strcmp (m->name, "kernel") == 0)
+    return startLoc - s->sec_load_offset + s->static_addr;
+
   /* eh_frame data has been loaded in the kernel, so readjust offset. */
   if (is_ehframe) {
     dbug_unwind(2, "eh_frame=%lx, eh_frame_addr=%lx\n", (unsigned long) m->eh_frame, m->eh_frame_addr);
@@ -768,7 +780,7 @@ static u32 *_stp_search_unwind_hdr(unsigned long pc,
 	unsigned long eh_hdr_addr = m->unwind_hdr_addr;
 
 	if (hdr == NULL || hdr_len < 4 || hdr[0] != 1) {
-		_stp_warn("no or bad debug frame hdr\n");
+		dbug_unwind(1, "no or bad debug frame hdr\n");
 		return NULL;
 	}
 
@@ -1107,8 +1119,8 @@ static int compute_expr(const u8 *expr, struct unwind_frame_info *frame,
 				unsigned long addr = POP;
 				switch (value) {
 #define CASE(n)     		case sizeof(u##n):			\
-					if (unlikely(_stp_read_address(value, (u##n *)addr, \
-						                       (user ? USER_DS : KERNEL_DS)))) \
+					if (unlikely(_stp_deref_nofault(value, sizeof(u##n), (u##n *)addr, \
+									(user ? USER_DS : KERNEL_DS)))) \
 						goto copy_failed;	\
 					break
 					CASES;
@@ -1130,7 +1142,7 @@ static int compute_expr(const u8 *expr, struct unwind_frame_info *frame,
 	return 0;
 
 copy_failed:
-	_stp_warn("_stp_read_address failed to access memory for deref\n");
+	_stp_warn("failed to access memory for deref\n");
 	return 1;
 truncated:
 	_stp_warn("invalid (truncated) DWARF expression in CFI\n");
@@ -1169,7 +1181,7 @@ static int unwind_frame(struct unwind_context *context,
 	unsigned i;
 	signed ptrType = -1, call_frame = 1;
 	uleb128_t retAddrReg = 0;
-	struct unwind_state *state = &context->state;
+	struct uw_state *state = &context->state;
 	unsigned long addr;
 
 	if (unlikely(table_len == 0)) {
@@ -1226,7 +1238,7 @@ static int unwind_frame(struct unwind_context *context,
 	       if it didn't exist. These should never be missing except
 	       when there are toolchain bugs. */
 	    unsigned long tableSize;
-	    _stp_warn("No binary search table for %s frame, doing slow linear search for %s\n", (is_ehframe ? "eh" : "debug"), m->path);
+	    dbug_unwind(1, "No binary search table for %s frame, doing slow linear search for %s\n", (is_ehframe ? "eh" : "debug"), m->path);
 	    for (fde = table, tableSize = table_len; cie = NULL, tableSize > sizeof(*fde)
 		 && tableSize - sizeof(*fde) >= *fde; tableSize -= sizeof(*fde) + *fde, fde += 1 + *fde / sizeof(*fde)) {
 			dbug_unwind(3, "fde=%lx tableSize=%d\n", (long)*fde, (int)tableSize);
@@ -1427,8 +1439,8 @@ static int unwind_frame(struct unwind_context *context,
 			   for 32-on-64 bit unwinding we need to ensure this is 0xFFFFFFFF */
 			switch (reg_info[i].width) {
 #define CASE(n)     case sizeof(u##n):					\
-				if (unlikely(_stp_read_address(FRAME_REG(i, u##n), (u##n *)addr, \
-							       (user ? USER_DS : KERNEL_DS)))) \
+				if (unlikely(_stp_deref_nofault(FRAME_REG(i, u##n), sizeof(u##n), (u##n *)addr, \
+								(user ? USER_DS : KERNEL_DS)))) \
 					goto copy_failed;		\
 				if (compat_task) FRAME_REG(i, u##n) &= 0xFFFFFFFF; \
 				dbug_unwind(1, "set register %d to %lx\n", i, (long)FRAME_REG(i,u##n)); \
@@ -1507,10 +1519,13 @@ static int unwind(struct unwind_context *context, int user)
           }
 
 	if (unlikely(m == NULL)) {
-                if (module_name)
-                        _stp_warn ("Missing unwind data for module, rerun with 'stap -d %s'\n",
-                                   module_name);
-		// Don't _stp_warn about this, will use fallback unwinder.
+                // some heuristics for the module name; we can't call
+                // kernel_text_address or friends from this context.
+                if (! module_name && (unsigned long)pc > PAGE_OFFSET)
+                        module_name = "kernel";
+                _stp_warn ("Missing unwind data for a module, rerun with 'stap -d %s'\n",
+                           module_name ?: "(unknown; retry with -DDEBUG_UNWIND)");
+		// Don't _stp_warn including the pc#, since it'll defeat warning deduplicator
 		dbug_unwind(1, "No module found for pc=%lx", pc);
 		return -EINVAL;
 	}
